{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE QuasiQuotes   #-}

module Gargantext.MicroServices.ReverseProxy (
  microServicesProxyApp
  ) where

import Prelude

import Conduit
import Data.ByteString qualified as B
import Data.ByteString.Builder
import Data.ByteString.Char8 qualified as C8
import Data.Conduit.List qualified as CC
import Data.String
import Data.Text qualified as T
import Data.Text.Encoding qualified as TE
import GHC.Generics
import Gargantext.API.Admin.EnvTypes
import Gargantext.API.Admin.Settings.MicroServices
import Gargantext.API.Admin.Types
import Gargantext.Database.Prelude (hasConfig)
import Gargantext.Prelude
import Gargantext.Prelude.Config (gc_frame_write_url)
import Network.HTTP.ReverseProxy
import Network.HTTP.Types (hCacheControl, RequestHeaders, hReferer, ResponseHeaders)
import Network.HTTP.Types.Header (hHost)
import Network.URI as URI
import Network.Wai (Request, rawPathInfo, requestHeaders)
import Servant
import Servant.Auth.Swagger ()
import Servant.Server.Generic
import Text.RE.TDFA.ByteString
import Text.RE.Replace hiding (Capture)

--
-- Types
--

data ProxyDestination =
  ProxyDestination
    { fwdProto :: C8.ByteString
    , fwdHost  :: C8.ByteString
    , fwdPort  :: PortNumber
    }

renderProxyDestination :: ProxyDestination -> C8.ByteString
renderProxyDestination ProxyDestination{..}
  = fwdProto <> "//" <> fwdHost <> ":" <> C8.pack (Prelude.show fwdPort)

--
-- The API
--

data ReverseProxyAPI mode = ReverseProxyAPI
  { notesProxy    :: mode :- NamedRoutes NotesProxy
  , notesSocketIo :: mode :- "socket.io"   :> NamedRoutes SocketIOProxy
  , meEndpoint    :: mode :- "me"          :> Raw
  } deriving Generic

data NotesProxy mode = NotesProxy
  { noteProxyEp :: mode :- Capture "frameId" T.Text :> Raw
  , slideEp     :: mode :- Capture "frameId" T.Text :> "slide" :> Raw
  } deriving Generic

data SocketIOProxy mode = SocketIOProxy
  { socketIoEp  :: mode :- QueryParam "noteId" T.Text :> Raw
  } deriving Generic

--
-- The Server
--

microServicesProxyApp :: Env -> Application
microServicesProxyApp env = genericServe (server env)

server :: Env -> ReverseProxyAPI AsServer
server env = ReverseProxyAPI {
    notesProxy    = notesProxyImplementation env
  , notesSocketIo = socketIOProxyImplementation env
  , meEndpoint    = proxyPassServer env
  }

notesProxyImplementation :: Env -> NotesProxy AsServer
notesProxyImplementation env = NotesProxy {
    noteProxyEp = \_frameId -> notesProxyServer env
  , slideEp     = \frameId  -> slideProxyServer env frameId
  }

socketIOProxyImplementation :: Env -> SocketIOProxy AsServer
socketIOProxyImplementation env = SocketIOProxy {
    socketIoEp = \_noteId -> proxyPassServer env
  }

notesProxyServer :: Env -> ServerT Raw m
notesProxyServer env = defaultForwardServer id env

slideProxyServer :: Env -> T.Text -> ServerT Raw m
slideProxyServer env frameId =
  defaultForwardServer (\rq -> rq { rawPathInfo = changePath (rawPathInfo rq) }) env
  where
    changePath :: ByteString -> ByteString
    changePath _ = TE.encodeUtf8 $ "/p/" <> frameId <> "#/"

-- Generic server forwarder
proxyPassServer :: Env -> ServerT Raw m
proxyPassServer env = defaultForwardServer id env


mkProxyDestination :: Env -> ProxyDestination
mkProxyDestination env = fromMaybe (panicTrace "Invalid URI found in the proxied Request.") $ do
  URI{..} <- URI.parseURI (T.unpack $ env ^. hasConfig . gc_frame_write_url)
  uriAut  <- uriAuthority
  port    <- case uriPort uriAut of
    "" -> pure 80
    xs -> readMaybe (dropWhile (==':') xs)
  pure $ ProxyDestination (C8.pack uriScheme) (C8.pack $ uriRegName uriAut) port

--
-- Combinators over the input Request
--

removeProxyPath :: T.Text -> Request -> Request
removeProxyPath pth originalRequest =
  originalRequest { rawPathInfo = removePath (rawPathInfo originalRequest) }
  where
    removePath :: ByteString -> ByteString
    removePath = TE.encodeUtf8 . T.replace pth "" . TE.decodeUtf8


defaultForwardServer :: (Request -> Request) -> Env -> ServerT Raw m
defaultForwardServer presendModifyRequest env =
  Tagged $ waiProxyToSettings forwardRequest (proxySettings $ mkProxyDestination env) (env ^. env_manager)
  where

    proxyDestination :: ProxyDestination
    proxyDestination = mkProxyDestination env

    microSrvSettings :: MicroServicesSettings
    microSrvSettings = env ^. env_settings . microservicesSettings

    proxyListeningPort :: Int
    proxyListeningPort = microSrvSettings ^. msProxyPort

    proxySettings :: ProxyDestination -> WaiProxySettings
    proxySettings proxyDestination =
      defaultWaiProxySettings {
          wpsProcessBody = \_req _res -> Just $ replaceRelativeLinks ("http://localhost:8009")
        , wpsModifyResponseHeaders = \_req _res -> tweakResponseHeaders
        }

    newHost :: ProxyDestination -> RequestHeaders -> RequestHeaders
    newHost hst hdrs = (hHost, renderProxyDestination hst) : filter ((/=) hHost . fst) hdrs

    newReferer :: RequestHeaders -> RequestHeaders
    newReferer hdrs =
      (hReferer, fromString $ "http://localhost:" <> Prelude.show proxyListeningPort) :
      filter ((/=) hHost . fst) hdrs

    -- | Forwards the request by substituting back the proxied address into the actual one.
    forwardRequest :: Request -> IO WaiProxyResponse
    forwardRequest originalRequest = do
      let proxiedReq = presendModifyRequest $ originalRequest {
                         requestHeaders = (newReferer . newHost proxyDestination . noCache $ (requestHeaders originalRequest))
                         }
      pure $ WPRModifiedRequest proxiedReq (ProxyDest (fwdHost proxyDestination) (fwdPort proxyDestination))

--
-- Utility functions
--

noCache :: RequestHeaders -> RequestHeaders
noCache hdrs = (hCacheControl, fromString "no-cache") : filter ((/=) hCacheControl . fst) hdrs

-- | Tweak the response headers so that they will have a bit more permissive
-- 'Content-Security-Policy'.
tweakResponseHeaders :: ResponseHeaders -> ResponseHeaders
tweakResponseHeaders = Prelude.map tweakHeader
  where
    tweakHeader (k,v)
      | k == "Content-Security-Policy"
      = (k, fromString "default-src *; style-src * 'unsafe-inline'; script-src * 'unsafe-inline' 'unsafe-eval'; img-src * data: 'unsafe-inline'; connect-src * 'unsafe-inline'; frame-src *;")
      | otherwise
      = (k,v)

-- | Replaces the relative links in any HTML blob returned by the proxy.
replaceRelativeLinks :: B.ByteString -> ConduitT B.ByteString (Flush Builder) IO ()
replaceRelativeLinks assetPath = CC.map flushReplace
  where
    --  Replaces the relative links in the proxied page content with proper urls.
    flushReplace :: B.ByteString -> Flush Builder
    flushReplace = Chunk . byteString . replaceIt

    replaceIt :: B.ByteString -> B.ByteString
    replaceIt htmlBlob =
      replaceAllCaptures ALL makeAbsolute $ htmlBlob *=~ [re|src="\/build\/|href="\/build\/|src="\/config|src="\/js\/|]
      where
        makeAbsolute _ _loc cap = case capturedText cap of
          "src=\"/build/"  -> Just $ "src=\"" <> assetPath <> "/build/"
          "href=\"/build/" -> Just $ "href=\"" <> assetPath <> "/build/"
          "src=\"/config"  -> Just $ "src=\"" <> assetPath <> "/config"
          "src=\"/js/"     -> Just $ "src=\"" <> assetPath <> "/js/"
          _                -> Just $ assetPath <> capturedText cap
